#include "vdmgr.h"
#include "ym2413.h"
#include "emu2413/emu2413.h"

	// YM2413
	typedef struct {
		VDMGR						*p_vdmgr ;						// zfoCXǗ
		IVDMOD						vdmod ;							// W[C^tF[X
		IVDMODPROPERTY				property ;						// vpeB[C^tF[X
		IVDMODADDRESSBUS			addressBus ;					// AhXoXC^tF[X
		IVDMODSOUNDGEN				soundGen ;						// TEhWFl[^C^tF[X
		OPLL						*p_opll ;						// OPLL AhX
		vdmgr_u32					baseClock ;						// x[XNbN
		vdmgr_u8					address ;						// AhX
		vdmgr_u8					type ;							// 
		vdmgr_u16					sampleFreq ;					// TvOg
	} VDMODYM2413 ;


// -------------------------------------------------------------------------------- 
//                                                                              
//                                                                                  
//   IN : p_this = this |C^                                                    
//   OUT: Ȃ                                                                      
// -------------------------------------------------------------------------------- 
static void VDMGR_METHODCALL module_release( IVDMOD *p_this )
{
	VDMODYM2413		*p=VDMGR_THISPTR(VDMODYM2413, vdmod, p_this) ;


	if( p->p_opll != NULL ) {
		OPLL_delete( p->p_opll ) ;
	}

	vdlib_free( p ) ;
}


// -------------------------------------------------------------------------------- 
//   Rg[                                                                   
//                                                                                  
//   IN : p_this  = this |C^                                                   
//        control = Rg[ԍ                                                
//        param   = Pp[^                                                  
//        p_param = Qp[^                                                  
//   OUT: VDMGR_S_OK                                                                
//        VDMGR_E_INVALIDPARAM                                                      
//        VDMGR_E_INVALIDCALL                                                       
// -------------------------------------------------------------------------------- 
static vdmgr_int VDMGR_METHODCALL module_control(IVDMOD *p_this, vdmgr_uint control, vdmgr_u32 param, void *p_param)
{
	VDMODYM2413		*p=VDMGR_THISPTR(VDMODYM2413, vdmod, p_this) ;


	switch( control ) {
		case VDCTRL_BUILDCOMPLETE:
		{
			p->p_opll = OPLL_new(p->baseClock, p->sampleFreq) ;
			if( p->p_opll == NULL ) {
				return VDMGR_E_NOTENOUGHMEMORY ;
			}

			OPLL_setChipType(p->p_opll, p->type) ;
			OPLL_resetPatch(p->p_opll, p->type) ;
			break ;
		}

		case VDCTRL_RESET:
		{
			OPLL_reset( p->p_opll ) ;
			break ;
		}

		case VDCTRL_SETSOUNDFORMAT:
		{
			p->sampleFreq = (vdmgr_u16 )param ;

			OPLL_setRate(p->p_opll, p->sampleFreq) ;
			break ;
		}

		case VDCTRL_GETADDRESSBUS: { *(IVDMODADDRESSBUS **)p_param = &p->addressBus ;  break ; }
		case VDCTRL_GETSOUNDGEN:   { *(IVDMODSOUNDGEN   **)p_param = &p->soundGen ;    break ; }
		case VDCTRL_GETPROPERTY:   { *(IVDMODPROPERTY   **)p_param = &p->property ;    break ; }

		default:
			return VDMGR_E_INVALIDCALL ;
	}

	return VDMGR_S_OK ;
}


// -------------------------------------------------------------------------------- 
//   擾                                                                           
//                                                                                  
//   IN : p_this = this |C^                                                    
//        p_prop = vpeB                                                       
//   OUT: VDMGR_S_OK                                                                
//        VDMGR_E_INVALIDCALL                                                       
// -------------------------------------------------------------------------------- 
static vdmgr_int VDMGR_METHODCALL property_get(IVDMODPROPERTY *p_this, VDMODPROP *p_prop)
{
	VDMODYM2413		*p=VDMGR_THISPTR(VDMODYM2413, property, p_this) ;


	p_prop->type = VDMODPROP_TYPE_U32 ;
	p_prop->u32  = p->type ;

	return VDMGR_S_OK ;
}


// -------------------------------------------------------------------------------- 
//   ݒ                                                                           
//                                                                                  
//   IN : p_this = this |C^                                                    
//        p_prop = vpeB                                                       
//   OUT: VDMGR_S_OK                                                                
//        VDMGR_E_INVALIDPARAM                                                      
//        VDMGR_E_INVALIDCALL                                                       
// -------------------------------------------------------------------------------- 
vdmgr_int VDMGR_METHODCALL property_set(IVDMODPROPERTY *p_this, VDMODPROP *p_prop)
{
	VDMODYM2413		*p=VDMGR_THISPTR(VDMODYM2413, property, p_this) ;


	if( p_prop->u32 > OPLL_281B_TONE ) {
		return VDMGR_E_INVALIDPARAM ;
	}

	if( p->type != (vdmgr_u8 )p_prop->u32 ) {
		p->type = (vdmgr_u8 )p_prop->u32 ;
		if( p->p_opll != NULL ) {
			OPLL_setChipType(p->p_opll, p->type) ;
			OPLL_resetPatch(p->p_opll, p->type) ;
			OPLL_reset( p->p_opll ) ;
		}
	}

	return VDMGR_S_OK ;
}


// -------------------------------------------------------------------------------- 
//   8bit                                                                   
//                                                                                  
//   IN : p_this  = this |C^                                                   
//        address = AhX                                                        
//        value   = l                                                              
//   OUT: Ȃ                                                                      
// -------------------------------------------------------------------------------- 
static void VDMGR_METHODCALL addressbus_write8(IVDMODADDRESSBUS *p_this, vdmgr_u32 address, vdmgr_u8 value)
{
	VDMODYM2413		*p=VDMGR_THISPTR(VDMODYM2413, addressBus, p_this) ;


	if( !(address&1) ) {
		p->address = value ;
	} else {
		OPLL_writeReg(p->p_opll, p->address, value) ;
	}
}


// -------------------------------------------------------------------------------- 
//   16bit                                                                  
//                                                                                  
//   IN : p_this  = this |C^                                                   
//        address = AhX                                                        
//        value   = l                                                              
//   OUT: Ȃ                                                                      
// -------------------------------------------------------------------------------- 
static void VDMGR_METHODCALL addressbus_write16(IVDMODADDRESSBUS *p_this, vdmgr_u32 address, vdmgr_u16 value)
{
	addressbus_write8(p_this, address, (vdmgr_u8 )value) ;
}


// -------------------------------------------------------------------------------- 
//   32bit                                                                  
//                                                                                  
//   IN : p_this  = this |C^                                                   
//        address = AhX                                                        
//        value   = l                                                              
//   OUT: Ȃ                                                                      
// -------------------------------------------------------------------------------- 
static void VDMGR_METHODCALL addressbus_write32(IVDMODADDRESSBUS *p_this, vdmgr_u32 address, vdmgr_u32 value)
{
	addressbus_write8(p_this, address, (vdmgr_u8 )value) ;
}


// -------------------------------------------------------------------------------- 
//   擾                                                                           
//                                                                                  
//   IN : p_this = this |C^                                                    
//        p_data = f[^AhX                                                   
//   OUT: Ȃ                                                                      
// -------------------------------------------------------------------------------- 
static void VDMGR_METHODCALL soundgen_getData(IVDMODSOUNDGEN *p_this, vdmgr_s16 *p_data)
{
	VDMODYM2413		*p=VDMGR_THISPTR(VDMODYM2413, soundGen, p_this) ;


	p_data[0] = p_data[1] = OPLL_calc( p->p_opll ) ;
}


// -------------------------------------------------------------------------------- 
//   W[擾                                                           
//                                                                                  
//   IN : p_vdmodInfo = W[i[AhX                                  
//   OUT: Ȃ                                                                      
// -------------------------------------------------------------------------------- 
void vdmod_ym2413_getModuleInfo( VDMODINFO *p_vdmodInfo )
{
	static const VDMODITEM	item[]={
								{ VDMGR_WS("type"),       VDMODITEM_TYPE_PROPERTY,   VDMODPROPERTY_SPEC_GET|VDMODPROP_TYPE_U32,                                    0 },
								{ VDMGR_WS("addressbus"), VDMODITEM_TYPE_ADDRESSBUS, VDMODADDRESSBUS_SPEC_GET|VDMODADDRESSBUS_SPEC_BUS|VDMODADDRESSBUS_SPEC_WRITE, 0 },
								{ VDMGR_WS("soundgen"),   VDMODITEM_TYPE_SOUNDGEN,   VDMODSOUNDGEN_SPEC_GET,                                                       0 }
							} ;


	p_vdmodInfo->p_name         = VDMGR_WS("ym2413") ;
	p_vdmodInfo->version        = 0x0000000a ;
	p_vdmodInfo->type           = VDMODINFO_TYPE_GENERAL ;
	p_vdmodInfo->p_item         = &item[0] ;
	p_vdmodInfo->itemCount      = VDMGR_COUNTOF( item ) ;
	p_vdmodInfo->create.general = vdmod_ym2413_create ;
}


// -------------------------------------------------------------------------------- 
//   W[쐬                                                               
//                                                                                  
//   IN : p_vdmgr     = zfoCXǗ                                            
//        param       = p[^                                                  
//   OUT: VDMGR_NULL != W[̃C^tF[X                                  
//        VDMGR_NULL == ُI                                                    
// -------------------------------------------------------------------------------- 
IVDMOD *vdmod_ym2413_create(VDMGR *p_vdmgr, vdmgr_u32 param)
{
	VDMODYM2413		*p ;


	p = vdlib_zalloc( sizeof( VDMODYM2413 ) ) ;
	if( p == VDMGR_NULL ) {
		return VDMGR_NULL ;
	}
	p->p_vdmgr            = p_vdmgr ;
	p->vdmod.release      = module_release ;
	p->vdmod.control      = module_control ;
	p->property.get       = property_get ;
	p->property.set       = property_set ;
	p->addressBus.read    = vdlib_addressbus_read ;
	p->addressBus.write8  = addressbus_write8 ;
	p->addressBus.write16 = addressbus_write16 ;
	p->addressBus.write32 = addressbus_write32 ;
	p->soundGen.getData   = soundgen_getData ;
	p->sampleFreq         = VDMODSOUNDGEN_SAMPLE_INIT ;

	p->baseClock = param ;
	if( !param ) {
		p->baseClock = 3600000 ;
	}

	return &p->vdmod ;
}
